from typing import Union
import os
import shutil
import json
from .get_kachery_dir import get_kachery_dir
from ._compute_file_hash import _compute_file_hash


def load_file_local(uri: str, *, dest: Union[None, str] = None) -> Union[str, None]:
    query = _get_query_from_uri(uri)
    assert uri.startswith("sha1://"), f"Invalid local URI: {uri}"
    a = uri.split("?")[0].split("/")
    assert len(a) >= 3, f"Invalid or unsupported URI: {uri}"
    sha1 = a[2]

    kachery_dir = get_kachery_dir()

    s = sha1
    parent_dir = f"{kachery_dir}/sha1/{s[0]}{s[1]}/{s[2]}{s[3]}/{s[4]}{s[5]}"
    filename = f"{parent_dir}/{sha1}"
    if os.path.exists(filename):
        if dest is not None:
            shutil.copyfile(filename, dest)
            return dest
        return filename

    if "location" in query:
        location = query["location"]
        if os.path.isabs(location) and os.path.exists(location):
            sha1_2 = _compute_file_hash(location, "sha1")
            if sha1_2 == sha1:
                if dest is not None:
                    shutil.copyfile(location, dest)
                    return dest
                return location

    # check for linked file
    s = sha1
    linked_file_record_parent_dir = (
        f"{kachery_dir}/linked_files/sha1/{s[0]}{s[1]}/{s[2]}{s[3]}/{s[4]}{s[5]}"
    )
    linked_file_record_path = f"{linked_file_record_parent_dir}/{sha1}"
    if os.path.exists(linked_file_record_path):
        with open(linked_file_record_path, "r") as f:
            a_txt = f.read()
        a = json.loads(a_txt)
        path0 = a["path"]
        size0 = a["size"]
        mtime0 = a["mtime"]
        if os.path.exists(path0):
            if os.path.getsize(path0) == size0 and os.stat(path0).st_mtime == mtime0:
                if dest is not None:
                    shutil.copyfile(path0, dest)
                    return dest
                return path0
            if (os.path.getsize(path0) == size0) and (
                _compute_file_hash(path0, algorithm="sha1") == sha1
            ):
                # file mtime has been updated, but hash is still the same
                with open(linked_file_record_path, "w") as f:
                    f.write(
                        json.dumps(
                            {
                                "path": path0,
                                "size": os.path.getsize(path0),
                                "mtime": os.stat(path0).st_mtime,
                                "sha1": sha1,
                            }
                        )
                    )
                if dest is not None:
                    shutil.copyfile(path0, dest)
                    return dest
                return path0
            else:
                print(f"Warning: sha1 of linked file has changed: {path0} {uri}")

    return None


def _get_query_from_uri(uri: str):
    a = uri.split("?")
    ret = {}
    if len(a) < 2:
        return ret
    b = a[1].split("&")
    for c in b:
        d = c.split("=")
        if len(d) == 2:
            ret[d[0]] = d[1]
    return ret



from typing import Optional
import os
import shutil
from ._load_http_file import _load_http_file
from ._load_github_file import _load_github_file
from .load_file_local import load_file_local
from ._load_sha1_file_from_cloud import _load_sha1_file_from_cloud


def load_file(
    uri: str,
    *,
    dest: Optional[str] = None,
    local_only: bool = False,
    _get_info: bool = False,
    verbose: bool = False
) -> Optional[str]:
    if uri.startswith("gh://"):
        if _get_info:
            raise Exception("Cannot use _get_info for this uri")
        return _load_github_file(uri)

    if uri.startswith("http://") or uri.startswith("https://"):
        if _get_info:
            raise Exception("Cannot use _get_info for this uri")
        return _load_http_file(uri)

    if local_only:
        if _get_info:
            raise Exception("Cannot use _get_info with local_only")
        return load_file_local(uri, dest=dest)
    if uri.startswith("/"):
        if _get_info:
            raise Exception("Cannot use _get_info for this uri")
        if os.path.exists(uri):
            if dest is not None:
                shutil.copyfile(uri, dest)
                return dest
            return uri
        else:
            return None
    if uri.startswith("sha1://"):
        if not _get_info:
            x = load_file_local(uri, dest=dest)
            if x is not None:
                return x
        sha1 = uri.split("?")[0].split("/")[2]
        fname_or_info = _load_sha1_file_from_cloud(
            sha1, verbose=verbose, dest=dest, _get_info=_get_info
        )
        if fname_or_info is None:
            return None
        elif isinstance(fname_or_info, dict):
            if not _get_info:
                raise Exception("Unexpected dict in load_file")
            return fname_or_info  # type: ignore
        elif isinstance(fname_or_info, str):
            if _get_info:
                raise Exception("Unexpected string in load_file")
            return fname_or_info
        else:
            raise Exception("Unexpected return value from _load_sha1_file_from_cloud")


def load_file_info(uri: str) -> dict:
    return load_file(uri, _get_info=True)  # type: ignore



import os
import shutil
import requests
from typing import Union
from urllib.parse import quote
import time

from .get_kachery_dir import get_kachery_dir

from .store_file_local import _random_string
from .store_file_local import _compute_file_hash, store_file_local
from ._custom_storage_backend import get_custom_storage_backend

from ._api_requests import _initiate_file_upload_request, _finalize_file_upload_request
from ._fs_operations import _makedirs


def store_file(
    filename: str,
    *,
    label: Union[str, None] = None,
    cache_locally: bool = False,
    local: bool = False,
):
    if local:
        return store_file_local(filename, label=label)
    if os.environ.get("KACHERY_STORE_FILE_DIR") is not None:
        return store_file_local(
            filename,
            label=label,
            store_file_dir=os.environ["KACHERY_STORE_FILE_DIR"],
            store_file_prefix=os.getenv("KACHERY_STORE_FILE_PREFIX", None),
        )

    _custom_storage_backend = get_custom_storage_backend()
    use_custom_storage_backend = _custom_storage_backend is not None and hasattr(
        _custom_storage_backend, "store_file"
    )
    if not use_custom_storage_backend:
        size = os.path.getsize(filename)
        alg = "sha1"
        hash0 = _compute_file_hash(filename, algorithm=alg)
        uri = f"{alg}://{hash0}"
        kachery_zone = os.environ.get("KACHERY_ZONE", "default")
        timer = time.time()
        response = None
        while True:
            KACHERY_API_KEY = os.environ.get("KACHERY_API_KEY")
            if not KACHERY_API_KEY and kachery_zone != "scratch":
                print("")
                print(
                    """
Kachery let's scientist store data files in the cloud for the purpose of using
cloud-based visualization tools and collaborating with others. This is a free
service when used for scientific research purposes. In order to use it, you must
register using your GitHub account, provide your email, and briefly describe the
purpose of the research. To register, visit https://kachery.vercel.app. Then set
the KACHERY_API_KEY environment variable to your API key.

Alternatively, you can use the "scratch" zone which is subject to regular deletion
of files by setting the KACHERY_ZONE environment variable to "scratch".

For more information, visit https://github.com/magland/kachery.
""".strip()
                )
                print("")
                # wait until the user presses enter
                input("Press Enter to continue...")
                raise Exception("KACHERY_API_KEY environment variable is not set")

            response = _initiate_file_upload_request(
                size=size, hash_alg=alg, hash=hash0, zone=kachery_zone
            )
            already_exists = response.get("alreadyExists", False)
            already_pending = response.get("alreadyPending", False)
            if already_exists:
                if label is not None:
                    uri = f"{uri}?label={quote(label)}"
                return uri
            elif already_pending:
                elapsed = time.time() - timer
                if elapsed > 60:
                    raise Exception(f"Upload is already pending... timeout: {uri}")
                print(f"Upload is already pending... waiting to retry {uri}")
                time.sleep(5)
            else:
                break

        assert response is not None

        signed_upload_url = response["signedUploadUrl"]
        object_key = response["objectKey"]
        with open(filename, "rb") as f:
            resp_upload = requests.put(signed_upload_url, data=f)
            if resp_upload.status_code != 200:
                print(signed_upload_url)
                raise Exception(
                    f"Error uploading file to bucket ({resp_upload.status_code}) {resp_upload.reason}: {resp_upload.text}"
                )
        response2 = _finalize_file_upload_request(
            object_key=object_key,
            hash_alg=alg,
            hash0=hash0,
            kachery_zone=kachery_zone,
            size=size,
        )

        if not response2.get("success", False):
            raise Exception(f"Error finalizing file upload: {uri}")

        if label is not None:
            uri = f"{uri}?label={quote(label)}"

        if label is not None:
            uri = f"{uri}?label={quote(label)}"
    else:
        # custom storage backend
        assert _custom_storage_backend is not None
        uri = _custom_storage_backend.store_file(filename, label=label)
        hash0 = None  # only computed if needed

    if cache_locally:
        kachery_dir = get_kachery_dir()
        if hash0 is None:
            # this would be None for custom storage backend
            hash0 = _compute_file_hash(filename, algorithm="sha1")
        e = hash0
        cache_parent_dir = f"{kachery_dir}/hash0/{e[0]}{e[1]}/{e[2]}{e[3]}/{e[4]}{e[5]}"
        if not os.path.exists(cache_parent_dir):
            _makedirs(cache_parent_dir)
        cache_filename = f"{cache_parent_dir}/{hash0}"
        if not os.path.exists(cache_filename):
            tmp_filename = f"{cache_filename}.tmp.{_random_string(8)}"
            shutil.copyfile(filename, tmp_filename)
            try:
                os.rename(tmp_filename, filename)
                # _chmod_file(filename)
            except:  # noqa
                if not os.path.exists(cache_filename):
                    raise Exception(f"Problem renaming file: {tmp_filename} {filename}")

    return uri



import os
import requests
from ._sha1_of_string import _sha1_of_string


def _initiate_file_upload_request(
    *, size: int, hash_alg: str, hash: str, zone: str
) -> dict:
    KACHERY_API_KEY = os.environ.get("KACHERY_API_KEY")
    if not KACHERY_API_KEY and zone != "scratch":
        raise Exception("KACHERY_API_KEY environment variable is not set")
    # difficulty is hard-coded at 13 for now. It's around 15 milliseconds of work.
    work_token = _create_work_token(hash, difficulty=13)
    payload = {
        "type": "initiateFileUploadRequest",
        "size": size,
        "hashAlg": hash_alg,
        "hash": hash,
        "zoneName": zone,
        "workToken": work_token,
    }
    url = "https://kachery.vercel.app/api/initiateFileUpload"
    headers = {"Content-Type": "application/json"}
    if KACHERY_API_KEY:
        headers["Authorization"] = f"Bearer {KACHERY_API_KEY}"  # type: ignore
    resp = requests.post(url, json=payload, headers=headers)
    if resp.status_code != 200:
        raise Exception(
            f'Error in {payload["type"]} ({resp.status_code}) {resp.reason}: {resp.text}'
        )
    response = resp.json()
    return response


def _finalize_file_upload_request(
    *, object_key: str, hash_alg: str, hash0: str, kachery_zone: str, size: int
) -> dict:
    KACHERY_API_KEY = os.environ.get("KACHERY_API_KEY")
    if not KACHERY_API_KEY and kachery_zone != "scratch":
        raise Exception("KACHERY_API_KEY environment variable is not set")
    payload = {
        "type": "finalizeFileUploadRequest",
        "objectKey": object_key,
        "hashAlg": hash_alg,
        "hash": hash0,
        "zoneName": kachery_zone,
        "size": size,
    }
    url = "https://kachery.vercel.app/api/finalizeFileUpload"
    headers = {"Content-Type": "application/json"}
    if KACHERY_API_KEY:
        headers["Authorization"] = f"Bearer {KACHERY_API_KEY}"  # type: ignore
    resp = requests.post(url, json=payload, headers=headers)
    if resp.status_code != 200:
        raise Exception(
            f'Error in {payload["type"]} ({resp.status_code}) {resp.reason}: {resp.text}'
        )
    response = resp.json()
    return response


def find_file_request(*, hash_alg: str, hash: str, zone: str) -> dict:
    KACHERY_API_KEY = os.environ.get("KACHERY_API_KEY")
    payload = {
        "type": "findFileRequest",
        "hashAlg": hash_alg,
        "hash": hash,
        "zoneName": zone,
    }
    url = "https://kachery.vercel.app/api/findFile"
    headers = {
        "Content-Type": "application/json",
    }
    if KACHERY_API_KEY:
        headers["Authorization"] = f"Bearer {KACHERY_API_KEY}"
    resp = requests.post(url, json=payload, headers=headers)
    if resp.status_code != 200:
        raise Exception(
            f'Error in {payload["type"]} ({resp.status_code}) {resp.reason}: {resp.text}'
        )
    response = resp.json()
    return response


def _create_work_token(hash: str, *, difficulty: int) -> str:
    while True:
        work_token = os.urandom(8).hex()
        if _check_work_token(work_token, hash, difficulty=difficulty):
            return work_token


def _check_work_token(work_token: str, hash: str, *, difficulty: int) -> bool:
    bits = _sha1_bits(hash + work_token)
    prefix = "0" * difficulty
    return bits.startswith(prefix)


def _sha1_bits(input: str) -> str:
    sha1 = _sha1_of_string(input)
    # covert to binary string of 0s and 1s
    return bin(int(sha1, 16))[2:].zfill(160)



import os
import click
import kachery as k2


@click.group(help="Kachery command-line client")
def cli():
    pass


@click.command(help="Store file in kachery")
@click.argument("filename")
@click.option("--cache-locally", is_flag=True)
@click.option("--label", required=False, default="")
def store_file(filename: str, cache_locally: bool, label: str):
    if label == "":
        label = os.path.basename(filename)
    uri = k2.store_file(filename, label=label, cache_locally=cache_locally)
    print(uri)


@click.command(help="Load file from kachery")
@click.argument("uri")
@click.option("--dest", required=False, default="")
def load_file(uri: str, dest: str):
    print(f"Loading file: {uri}")
    fname = k2.load_file(uri, dest=dest if len(dest) > 0 else None)
    if fname is not None:
        print(fname)


@click.command(help="Load file info from kachery")
@click.argument("uri")
def load_file_info(uri: str):
    import json

    x = k2.load_file_info(uri)
    if x is None:
        return
    print(json.dumps(x, indent=4))


@click.command(help="Load file from kachery and write to stdout")
@click.argument("uri")
def cat_file(uri: str):
    k2.cat_file(uri)


@click.command(help="Store file in the local cache")
@click.argument("filename")
@click.option("--label", required=False, default="")
def store_file_local(filename: str, label: str):
    if label == "":
        label = os.path.basename(filename)
    uri = k2.store_file_local(filename, label=label)
    print(uri)


cli.add_command(store_file)
cli.add_command(load_file)
cli.add_command(load_file_info)
cli.add_command(cat_file)
cli.add_command(store_file_local)



from typing import Any, Union
from tempfile import TemporaryDirectory

from .store_file import store_file
from .load_file import load_file
from .store_file_local import store_file_local


def store_text(
    text: str,
    *,
    label: Union[str, None] = None,
    cache_locally: bool = False,
    local: bool = False,
) -> str:
    with TemporaryDirectory() as tmpdir:
        fname = f"{tmpdir}/file.dat"
        with open(fname, "w") as f:
            f.write(text)
        return store_file(fname, label=label, cache_locally=cache_locally, local=local)


def store_json(
    x: Any,
    *,
    separators=(",", ":"),
    indent=None,
    label: Union[str, None] = None,
    cache_locally: bool = False,
    local: bool = False,
) -> str:
    import simplejson

    text = simplejson.dumps(
        x, separators=separators, indent=indent, allow_nan=False, sort_keys=True
    )
    return store_text(text, label=label, cache_locally=cache_locally, local=local)


def store_npy(
    array: Any,
    *,
    label: Union[str, None] = None,
    cache_locally: bool = False,
    local: bool = False,
) -> str:
    import numpy as np

    with TemporaryDirectory() as tmpdir:
        fname = f"{tmpdir}/file.npy"
        np.save(fname, array, allow_pickle=False)
        return store_file(fname, label=label, cache_locally=cache_locally, local=local)


def load_text(uri: str, *, local_only: bool = False) -> Union[str, None]:
    local_path = load_file(uri, local_only=local_only)
    if local_path is None:
        return None
    with open(local_path, "r") as f:
        return f.read()


def load_json(uri: str, *, local_only: bool = False) -> Union[dict, None]:
    import simplejson

    local_path = load_file(uri, local_only=local_only)
    if local_path is None:
        return None
    with open(local_path, "r") as f:
        return simplejson.load(f)


def load_npy(uri: str, *, local_only: bool = False) -> Union[Any, None]:
    import numpy as np

    local_path = load_file(uri, local_only=local_only)
    if local_path is None:
        return None
    return np.load(local_path, allow_pickle=False)


def store_text_local(text: str, label: Union[str, None] = None) -> str:
    with TemporaryDirectory() as tmpdir:
        fname = f"{tmpdir}/file.dat"
        with open(fname, "w") as f:
            f.write(text)
        return store_file_local(fname, label=label)


def store_json_local(
    x: Any, *, separators=(",", ":"), indent=None, label: Union[str, None] = None
) -> str:
    import simplejson

    text = simplejson.dumps(
        x, separators=separators, indent=indent, allow_nan=False, sort_keys=True
    )
    return store_text_local(text, label=label)


def store_npy_local(array: Any, label: Union[str, None] = None) -> str:
    import numpy as np

    with TemporaryDirectory() as tmpdir:
        fname = f"{tmpdir}/file.npy"
        np.save(fname, array, allow_pickle=False)
        return store_file_local(fname, label=label)



import os
import random
import atexit
import shutil


def use_sandbox(use: bool = True):
    if use:
        os.environ["KACHERY_USE_SANDBOX"] = "1"
        if not os.path.exists(os.environ["KACHERY_SANDBOX_DIR"]):
            # only create the dir when we are actually switching to sandbox mode
            os.makedirs(os.environ["KACHERY_SANDBOX_DIR"])
            # important to do it this way so we don't clean up the temp dir in a subprocesss
            _global["sandbox_dir_to_cleanup"] = os.environ["KACHERY_SANDBOX_DIR"]  # type: ignore
    else:
        # We turn off the sandbox, but if we turn it back on again, we want to use the same dir for this process
        os.environ["KACHERY_USE_SANDBOX"] = "0"


def get_kachery_dir(*, respect_sandbox: bool = True):
    from pathlib import Path

    if respect_sandbox and (os.getenv("KACHERY_USE_SANDBOX", "") == "1"):
        return os.environ["KACHERY_SANDBOX_DIR"]
    homedir = str(Path.home())
    hsd = os.getenv("KACHERY_DIR", f"{homedir}/.kachery")
    if not os.path.exists(hsd):
        os.makedirs(hsd)
    return hsd


def _random_string(num_chars: int) -> str:
    chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    return "".join(random.choice(chars) for _ in range(num_chars))


# We use environ variables for this so that the info is passed on to subprocesses
if os.getenv("KACHERY_SANDBOX_DIR", None) is None:
    # We don't actually mkdir until it is used
    # I found that electron remaps the /tmp directory, so we can't use /tmp. Probably just as well.
    os.environ["KACHERY_SANDBOX_DIR"] = (
        f"{get_kachery_dir()}/sandbox/{_random_string(8)}"
    )
if os.getenv("KACHERY_USE_SANDBOX", None) is None:
    os.environ["KACHERY_USE_SANDBOX"] = "0"

_global = {"sandbox_dir_to_cleanup": None}


def at_exit():
    # important to do it this way so we don't clean up the temp dir in a subprocesss
    sandbox_dir_to_cleanup = _global["sandbox_dir_to_cleanup"]
    if sandbox_dir_to_cleanup is not None:
        if os.path.exists(sandbox_dir_to_cleanup):
            print(f"Cleaning up sandbox dir: {sandbox_dir_to_cleanup}")
            shutil.rmtree(sandbox_dir_to_cleanup)


atexit.register(at_exit)



import os
import json
import shutil
from tempfile import TemporaryDirectory
from urllib.parse import quote
from typing import Union
from .get_kachery_dir import get_kachery_dir
from ._compute_file_hash import _compute_file_hash
from ._random_string import _random_string
from ._fs_operations import _makedirs


def store_file_local(
    filename: str,
    *,
    label: Union[str, None] = None,
    store_file_dir: Union[str, None] = None,
    store_file_prefix: Union[str, None] = None,
):
    from .load_file import load_file

    if filename.startswith("sha1://"):
        a = load_file(filename)
        if a is None:
            raise Exception(f"Problem loading file: {filename}")
        filename = a
    if not os.path.isabs(filename):
        filename = os.path.abspath(filename)
    sha1 = _compute_file_hash(filename, algorithm="sha1")
    if store_file_prefix is None:
        uri = f"sha1://{sha1}"
    else:
        uri = f"{store_file_prefix}/sha1/{sha1}"
    if label is not None:
        uri = f"{uri}?label={quote(label)}"
    if store_file_dir is None:
        kachery_dir = get_kachery_dir()
        kachery_storage_parent_dir = f"{kachery_dir}/sha1/{sha1[0]}{sha1[1]}/{sha1[2]}{sha1[3]}/{sha1[4]}{sha1[5]}"
        kachery_storage_file_name = f"{kachery_storage_parent_dir}/{sha1}"
    else:
        kachery_storage_parent_dir = f"{store_file_dir}/sha1"
        kachery_storage_file_name = f"{kachery_storage_parent_dir}/{sha1}"
    if not os.path.exists(kachery_storage_file_name):
        if not os.path.exists(kachery_storage_parent_dir):
            _makedirs(kachery_storage_parent_dir)
        tmp_filename = f"{kachery_storage_file_name}.{_random_string(10)}"
        shutil.copyfile(filename, tmp_filename)
        try:
            os.rename(tmp_filename, kachery_storage_file_name)
            # _chmod_file(kachery_storage_file_name)
        except Exception:
            # Maybe another client renamed the file
            if not os.path.exists(kachery_storage_file_name):
                raise Exception(
                    f"Unexpected problem renaming file: {tmp_filename} {kachery_storage_file_name}"
                )
    return uri


def store_json_local(obj, *, label: Union[str, None] = None):
    with TemporaryDirectory() as tmpdir:
        fname = f"{tmpdir}/tmp.json"
        with open(fname, "w") as f:
            json.dump(obj, f)
        return store_file_local(fname, label=label)



from typing import Union
import os
import shutil
import requests
from .get_kachery_dir import get_kachery_dir
from ._random_string import _random_string
from ._api_requests import find_file_request
from ._fs_operations import _makedirs


def _load_sha1_file_from_cloud(
    sha1: str, *, verbose: bool, dest: Union[None, str] = None, _get_info: bool = False
) -> Union[str, dict, None]:
    kachery_zone = os.environ.get("KACHERY_ZONE", "default")

    response = find_file_request(hash_alg="sha1", hash=sha1, zone=kachery_zone)

    found = response["found"]
    uri = f"sha1://{sha1}"
    if found:
        url = response["url"]
    else:
        return None

    if _get_info:
        # we don't want the user to get the idea they should use the URL directly!
        del response["url"]
        return response

    kachery_dir = get_kachery_dir()
    e = sha1
    parent_dir = f"{kachery_dir}/sha1/{e[0]}{e[1]}/{e[2]}{e[3]}/{e[4]}{e[5]}"
    filename = f"{parent_dir}/{sha1}"
    if verbose:
        print(f"Loading file from kachery: {uri}")
    if not os.path.exists(parent_dir):
        _makedirs(parent_dir)
    tmp_filename = f"{filename}.tmp.{_random_string(8)}"
    with requests.get(url, stream=True) as r:
        if r.status_code == 404:
            raise Exception(f"Unexpected: file not found in bucket: {url}")
        r.raise_for_status()
        with open(tmp_filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    try:
        os.rename(tmp_filename, filename)
        # _chmod_file(filename)
    except Exception:
        if not os.path.exists(filename):  # maybe some other process beat us to it
            raise Exception(f"Unexpected problem moving file {tmp_filename}")
    if dest is not None:
        shutil.copyfile(filename, dest)
        return dest
    return filename



import os
import hashlib
import json
from ._sha1_of_string import _sha1_of_string
from .mutable_local import get_mutable_local, set_mutable_local


def _compute_file_hash(path: str, algorithm: str) -> str:
    if not os.path.isabs(path):
        path = os.path.abspath(path)
    if not os.path.exists(path):
        raise Exception(f"File does not exist: {path}")
    size0 = os.path.getsize(path)
    if size0 > 1000 * 1000:
        a = get_mutable_local(f"@compute_sha1_cache/{_sha1_of_string(path)}")
        if a:
            a = json.loads(a)
            mtime = os.stat(path).st_mtime
            if a["size"] == size0 and a["mtime"] == mtime:
                return a["sha1"]
    if size0 > 1000 * 1000 * 100:
        print("Computing {} of {}".format(algorithm, path))
    BLOCKSIZE = 65536
    hashsum = getattr(hashlib, algorithm)()
    with open(path, "rb") as file:
        buf = file.read(BLOCKSIZE)
        while len(buf) > 0:
            hashsum.update(buf)
            buf = file.read(BLOCKSIZE)
    ret = hashsum.hexdigest()
    if size0 > 1000 * 1000:
        set_mutable_local(
            f"@compute_sha1_cache/{_sha1_of_string(path)}",
            json.dumps(
                {
                    "path": path,
                    "size": size0,
                    "mtime": os.stat(path).st_mtime,
                    "sha1": ret,
                }
            ),
        )
    return ret



import os
import stat


all_dir_mode = (
    stat.S_IRUSR
    | stat.S_IWUSR
    | stat.S_IXUSR
    | stat.S_IRGRP
    | stat.S_IWGRP
    | stat.S_IXGRP
    | stat.S_IROTH
    | stat.S_IWOTH
    | stat.S_IXOTH
)
all_file_mode = (
    stat.S_IRUSR
    | stat.S_IWUSR
    | stat.S_IRGRP
    | stat.S_IWGRP
    | stat.S_IROTH
    | stat.S_IWOTH
)


def _makedirs(path: str):
    # This is specific to L. Frank's multi-user setup where they have a shared directory for kachery
    multiuser = (os.getenv("KACHERY_MULTI_USER", "0") == "1") or (
        os.getenv("KACHERY_CLOUD_MULTI_USER", "0") == "1"
    )
    if multiuser:
        os.makedirs(path, mode=all_dir_mode)
    else:
        os.makedirs(path)
